/* eslint-disable no-restricted-syntax,no-continue */
import { dirname } from 'path';
import { warn } from '@linaria/logger';
import isRequire from './isRequire';

const safeResolve = (name, filename) => {
  try {
    return require.resolve(name, {
      paths: [dirname(filename)]
    });
  } catch (e) {
    return name;
  }
};

function getValue({
  node
}) {
  return node.type === 'Identifier' ? node.name : node.value;
} // We only need value imports. Type imports can be ignored.


const isExportKindIsValue = p => !p.node.exportKind || p.node.exportKind === 'value'; // We only need value exports. Type exports can be ignored.


const isImportKindIsValue = p => !p.node.importKind || p.node.importKind === 'value'; // Force TypeScript to check, that we have implementation for every possible specifier


const collectors = {
  ImportSpecifier(path) {
    if (!isImportKindIsValue(path)) return [];
    const imported = getValue(path.get('imported'));
    const local = path.get('local');
    return [{
      imported,
      local
    }];
  },

  ImportDefaultSpecifier(path) {
    const local = path.get('local');
    return [{
      imported: '=',
      local
    }];
  },

  ImportNamespaceSpecifier(path) {
    const local = path.get('local');
    return [{
      imported: '*',
      local
    }];
  }

};

function collectFromImportDeclaration(path, state) {
  // If importKind is specified, and it's not a value, ignore that import
  if (!isImportKindIsValue(path)) return;
  const source = safeResolve(getValue(path.get('source')), state.filename);
  path.get('specifiers').forEach(specifier => {
    if (specifier.isImportSpecifier() && !isImportKindIsValue(specifier)) return;
    const collector = collectors[specifier.node.type];
    state.imports.push(...collector(specifier).map(item => ({ ...item,
      source
    })));
  });
}

function whatIsDestructed(path) {
  const destructedProps = [];
  const properties = path.get('properties');
  properties.forEach(property => {
    if (property.isObjectProperty()) {
      const key = property.get('key');
      if (!key.isIdentifier() && !key.isStringLiteral()) return;
      const what = getValue(key);
      const as = property.get('value');
      if (!as.isIdentifier()) return;
      destructedProps.push({
        what,
        as
      });
      return;
    } // If it is not an ObjectProperty, it is a RestElement
    // In such case it is the same as importing the whole namespace


    if (property.isRestElement()) {
      const arg = property.get('argument');

      if (!arg.isIdentifier()) {
        // some unexpected type of argument
        return;
      }

      destructedProps.push({
        as: arg,
        what: '*'
      });
      return;
    }

    warn('evaluator:collectExportsAndImports', 'Unknown type of property:', property.node.type);
  });
  return destructedProps;
}

function fromVariableDeclarator(path, isSync) {
  const id = path.get('id');

  if (id.isIdentifier()) {
    // It's the simplest case when the full namespace is imported
    return [{
      as: id,
      what: '*'
    }];
  }

  if (!isSync) {
    // Something went wrong
    // Is it something like `const { … } = import(…)`?
    warn('evaluator:collectExportsAndImports', '`import` should be awaited');
    return [];
  }

  if (id.isObjectPattern()) {
    return whatIsDestructed(id);
  } // What else it can be?


  warn('evaluator:collectExportsAndImports', 'Unknown type of id', id.node.type);
  return [];
}

function collectFromDynamicImport(path, state) {
  const {
    parentPath: callExpression
  } = path;

  if (!callExpression.isCallExpression()) {
    // It's wrong `import`
    return;
  }

  const [sourcePath] = callExpression.get('arguments');

  if (!sourcePath || !sourcePath.isStringLiteral()) {
    // Import should have at least one argument, and it should be StringLiteral
    return;
  }

  const source = safeResolve(sourcePath.node.value, state.filename);
  let {
    parentPath: container,
    key
  } = callExpression;
  let isAwaited = false;

  if (container.isAwaitExpression()) {
    // If it's not awaited import, it imports the full namespace
    isAwaited = true;
    key = container.key;
    container = container.parentPath;
  } // Is it `const something = await import("something")`?


  if (key === 'init' && container.isVariableDeclarator()) {
    fromVariableDeclarator(container, isAwaited).map(prop => state.imports.push({
      imported: prop.what,
      local: prop.as,
      source
    }));
  }
}

function getImportTypeByInteropFunction(path) {
  const callee = path.get('callee');

  if (!callee.isIdentifier()) {
    return undefined;
  }

  const {
    name
  } = callee.node;

  if (name.startsWith('_interopRequireDefault')) {
    return '=';
  }

  if (name.startsWith('_interopRequireWildcard')) {
    return '*';
  }

  if (name.startsWith('__rest')) {
    return '*';
  }

  return undefined;
}

function collectFromRequire(path, state) {
  if (!isRequire(path)) return;
  const {
    parentPath: callExpression
  } = path;

  if (!callExpression.isCallExpression()) {
    // It's wrong `require`
    return;
  }

  const [sourcePath] = callExpression.get('arguments');

  if (!sourcePath || !sourcePath.isStringLiteral()) {
    // Import should have at least one argument, and it should be StringLiteral
    return;
  }

  const source = safeResolve(sourcePath.node.value, state.filename);
  const {
    parentPath: container,
    key
  } = callExpression;

  if (container.isCallExpression() && key === 0) {
    // It may be transpiled import such as
    // `var _atomic = _interopRequireDefault(require("@linaria/atomic"));`
    const imported = getImportTypeByInteropFunction(container);

    if (!imported) {
      // It's not a transpiled import.
      // TODO: Can we guess that it's a namespace import?
      warn('evaluator:collectExportsAndImports', 'Unknown wrapper of require', container.node.callee);
      return;
    }

    const {
      parentPath: variableDeclarator
    } = container;

    if (!variableDeclarator.isVariableDeclarator()) {
      // TODO: Where else it can be?
      warn('evaluator:collectExportsAndImports', 'Unexpected require inside', variableDeclarator.node.type);
      return;
    }

    const id = variableDeclarator.get('id');

    if (!id.isIdentifier()) {
      warn('evaluator:collectExportsAndImports', 'Id should be Identifier', variableDeclarator.node.type);
      return;
    }

    state.imports.push({
      imported,
      local: id,
      source
    });
  }

  if (container.isMemberExpression()) {
    // It is `require('@linaria/shaker').dep`
    const property = container.get('property');

    if (!property.isIdentifier() && !property.isStringLiteral()) {
      warn('evaluator:collectExportsAndImports', 'Property should be Identifier or StringLiteral', property.node.type);
      return;
    }

    const {
      parentPath: variableDeclarator
    } = container;

    if (variableDeclarator.isVariableDeclarator()) {
      // It is `const … = require('@linaria/shaker').dep`;
      const id = variableDeclarator.get('id');

      if (id.isIdentifier()) {
        state.imports.push({
          imported: getValue(property),
          local: id,
          source
        });
      } else {
        warn('evaluator:collectExportsAndImports', 'Id should be Identifier', variableDeclarator.node.type);
      }
    } else {
      // Maybe require is passed as an argument to some function?
      // Just use the whole MemberExpression as a local
      state.imports.push({
        imported: getValue(property),
        local: container,
        source
      });
    }

    return;
  } // Is it `const something = require("something")`?


  if (key === 'init' && container.isVariableDeclarator()) {
    fromVariableDeclarator(container, true).map(prop => state.imports.push({
      imported: prop.what,
      local: prop.as,
      source
    }));
  }
}

function unfoldNamespaceImports(imports) {
  const result = [];

  for (const importItem of imports) {
    if (importItem.imported === '*') {
      const {
        local
      } = importItem;

      if (!local.isIdentifier()) {
        // TODO: handle it
        result.push(importItem);
        continue;
      }

      const binding = local.scope.getBinding(local.node.name);

      if (!binding?.referenced) {
        // Imported namespace is not referenced and probably not used,
        // but it can have side effects, so we should keep it as is
        result.push(importItem);
        continue;
      }

      for (const referencePath of binding?.referencePaths ?? []) {
        const {
          parentPath
        } = referencePath;

        if (parentPath?.isMemberExpression() && referencePath.key === 'object') {
          const property = parentPath.get('property');

          if (property.isIdentifier()) {
            const {
              name
            } = property.node;
            result.push({ ...importItem,
              imported: name === 'default' ? '=' : name,
              local: parentPath // TODO: try to resolve identifier

            });
            continue;
          } // Otherwise, we can't predict usage and import it as is


          result.push(importItem);
          break;
        }

        if (parentPath?.isVariableDeclarator() && referencePath.key === 'init') {
          fromVariableDeclarator(parentPath, true).map(prop => result.push({ ...importItem,
            imported: prop.what,
            local: prop.as
          }));
          continue;
        }

        if (parentPath?.isExportSpecifier()) {
          // The whole namespace is re-exported
          result.push(importItem);
          break;
        } // Otherwise, we can't predict usage and import it as is
        // TODO: handle more cases


        warn('evaluator:unfoldNamespaceImports', 'Unknown reference', referencePath.node.type);
        result.push(importItem);
        break;
      }
    } else {
      result.push(importItem);
    }
  }

  return result;
}

function collectFromExportSpecifier(path, state) {
  if (path.isExportSpecifier()) {
    const local = path.get('local');
    const exported = getValue(path.get('exported'));
    state.exports.push({
      local,
      exported
    });
    return;
  } // TODO: handle other cases


  warn('evaluator:collectExportsAndImports', 'Unprocessed ExportSpecifier', path.node.type);
}

function collectFromExportNamedDeclaration(path, state) {
  if (!isExportKindIsValue(path)) return;
  const specifiers = path.get('specifiers');

  if (specifiers) {
    specifiers.forEach(specifier => collectFromExportSpecifier(specifier, state));
  }

  const declaration = path.get('declaration');

  if (declaration.isVariableDeclaration()) {
    declaration.get('declarations').forEach(declarator => {
      fromVariableDeclarator(declarator, true).forEach(prop => {
        // What is defined
        state.exports.push({
          exported: prop.what,
          local: prop.as
        });
      });
    });
  }
}

const cache = new WeakMap();
export default function collectExportsAndImports(path, filename) {
  const state = {
    imports: [],
    exports: [],
    filename
  };

  if (cache.has(path)) {
    return cache.get(path) ?? state;
  }

  path.traverse({
    ExportNamedDeclaration: collectFromExportNamedDeclaration,
    ImportDeclaration: collectFromImportDeclaration,
    Import: collectFromDynamicImport,
    Identifier: collectFromRequire
  }, state);
  state.imports = unfoldNamespaceImports(state.imports);
  cache.set(path, state);
  return state;
}
//# sourceMappingURL=collectExportsAndImports.js.map